// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

#pragma once

#include "opentelemetry/version.h"

OPENTELEMETRY_BEGIN_NAMESPACE
namespace nostd
{
template <typename... Ts>
class variant;
}  // namespace nostd
OPENTELEMETRY_END_NAMESPACE
