# Copyright The OpenTelemetry Authors
# SPDX-License-Identifier: Apache-2.0

# If comment markup is enabled, don't reflow the first comment block in
# eachlistfile. Use this to preserve formatting of your
# copyright/licensestatements.
first_comment_is_literal = True
